@@ -1,5 +1,5 @@
 # -*-Perl-*-
-# $Id: Sybase.pm,v 1.59 2004/06/16 13:44:54 mpeppler Exp $
+# $Id: Sybase.pm,v 1.85 2004/12/19 09:52:39 mpeppler Exp $
 
 # Copyright (c) 1996-2004   Michael Peppler
 #
@@ -25,8 +25,8 @@
 
     $hostname = Sys::Hostname::hostname();
     $init_done = 0;
-    $VERSION = '1.03';
-    my $Revision = substr(q$Revision: 1.59 $, 10);
+    $VERSION = '1.05';
+    my $Revision = substr(q$Revision: 1.85 $, 10);
 
     require_version DBI 1.30;
 
@@ -34,9 +34,9 @@
 
 
     $drh = undef;	# holds driver handle once initialised
-    $err = 0;		# The $DBI::err value
-    $errstr = '';
-    $sqlstate = "00000";
+ #   $err = 0;		# The $DBI::err value
+ #   $errstr = '';
+ #   $sqlstate = "00000";
 
     sub driver {
 	return $drh if $drh;
@@ -45,15 +45,22 @@
 	($drh) = DBI::_new_drh($class, {
 	    'Name' => 'Sybase',
 	    'Version' => $VERSION,
-	    'Err'     => \$DBD::Sybase::err,
-	    'Errstr'  => \$DBD::Sybase::errstr,
-	    'State'   => \$DBD::Sybase::sqlstate,
+#	    'Err'     => \$DBD::Sybase::err,
+#	    'Errstr'  => \$DBD::Sybase::errstr,
+#	    'State'   => \$DBD::Sybase::sqlstate,
 	    'Attribution' => 'Sybase DBD by Michael Peppler',
 	    });
 
-	if($DBI::VERSION >= 1.34 && !$DBD::Sybase::init_done) {
-	    #	*syb_nsql = *nsql;
+	if($DBI::VERSION >= 1.37 && !$DBD::Sybase::init_done) {
 	    DBD::Sybase::db->install_method('syb_nsql');
+	    DBD::Sybase::db->install_method('syb_date_fmt');
+	    DBD::Sybase::st->install_method('syb_ct_get_data');
+	    DBD::Sybase::st->install_method('syb_ct_data_info');
+	    DBD::Sybase::st->install_method('syb_ct_send_data');
+	    DBD::Sybase::st->install_method('syb_ct_prepare_send');
+	    DBD::Sybase::st->install_method('syb_ct_finish_send');
+	    DBD::Sybase::st->install_method('syb_output_params');
+	    DBD::Sybase::st->install_method('syb_describe');
 	    ++$DBD::Sybase::init_done;
 	}
 
@@ -84,39 +91,10 @@
 	    'CURRENT_USER' => $user,
 	    });
 
-	DBD::Sybase::db::_login($this, $server, $user, $auth, $attr) or return undef;
+	DBD::Sybase::db::_login($this, $server, $user, $auth, $attr) 
+		or return undef;
 
-	# OK - let us see what sort of server we're connected to. We
-	# need this because some of the SQL commands that we need to send
-	# to the server depend on this information.
-    {
-	
-	# local $this->{AutoCommit} = 1; #commented out - see bug #551.
-
-        # Attributes that are passed in only get set *after* this routine
-        # exits (i.e. when connect() returns).
-        # So we need to set any syb_xxx attributes here:
-        foreach (grep(/^syb/, keys(%$attr))) {
-	    $this->{$_} = $attr->{$_};
-	}
-	
-	my $sth = $this->prepare("select \@\@version");
-	if($sth->execute) {
-	    my $row = $sth->fetch;
-	    if($row && ref($row)) {
-		if($row->[0] =~ /microsoft/i) {
-		    $this->{syb_server_version} = -1;
-		    #warn "Connected to MS-SQL\n";
-		} elsif($row->[0] =~ /adaptive server enterprise\/([\d\.]+)/i ||
-		       $row->[0] =~ /sql server\/([\d\.]+)/i) {
-		    $this->{syb_server_version} = $1;
-		    #warn "Connected to Sybase\n";
-		}
-	    }
-	}
-    }
-
-	$this;
+	return $this;
     }
 
     sub data_sources {
@@ -199,18 +177,37 @@
 	$sth->execute(@params) or return undef;
 	return undef if $sth->err;
 	if(defined($sth->{syb_more_results})) {
-	    do {
+	    {
 		while(my $dat = $sth->fetch) {
 		    return undef if $sth->err;
 		    # XXX do something intelligent here...
 		}
-	    } while($sth->{syb_more_results});
+		redo if $sth->{syb_more_results};
+	    }
 	}
 	my $rows = $sth->rows;
 
 	($rows == 0) ? "0E0" : $rows;
     }
 
+    # This will only work if the statement handle used to do the insert
+    # has been properly freed. Otherwise this will try to fetch @@identity
+    # from a different (new!) connection - which is obviously wrong.
+    sub last_insert_id {
+	my ($dbh, $catalog, $schema, $table, $field, $attr) = @_;
+	# parameters are ignored.
+	
+	my $sth = $dbh->prepare('select @@identity');
+	if(!$sth->execute) {
+	    return undef;
+	}
+	my $value;
+	($value) = $sth->fetchrow_array;
+	$sth->finish;
+	
+	return $value;
+    }
+
     sub table_info {
 	my $dbh = shift;
 	my $catalog = $dbh->quote(shift);
@@ -294,6 +291,12 @@
 	$sth;
     }
 
+    sub begin_work {
+        my $dbh = shift;
+
+	return $dbh->STORE('AutoCommit', 0);
+    }
+
     sub ping {
 	my $dbh = shift;
 	return 0 if DBD::Sybase::db::_isdead($dbh);
@@ -408,7 +411,7 @@
 	    eval { 
 		$sth->execute || die $sth->errstr; 
 #		$sth->execute;
-		do {
+		{
 		    if ( $DBD::Sybase::syb_nsql_nostatus &&
 			 $sth->{syb_result_type} == &CS_STATUS_RESULT ) {
 			while ( $data = $sth->fetchrow_arrayref ) {
@@ -450,7 +453,8 @@
 		    
 		    die $sth->errstr if($sth->err);
 		    
-		} while($sth->{'syb_more_results'});
+		    redo if $sth->{syb_more_results};
+		}
 	    };
 	    # If $@ is set then something failed in the eval{} call above.
 	    if($@) {
@@ -493,7 +497,6 @@
     
     if($DBI::VERSION >= 1.34) {
 	*syb_nsql = *nsql;
-#	DBD::Sybase::db->install_method('syb_nsql');
     }
 }
 
@@ -507,7 +510,7 @@
 	my @results;
 	my $status;
 
-	do {
+	{
 	    while(my $d = $sth->fetch) {
 		# The tie() doesn't work here, so call the FETCH method
 		# directly....
@@ -517,7 +520,8 @@
 		   $status = $d->[0];
 	       }
 	    }
-	} while($sth->FETCH('syb_more_results'));
+	    redo if $sth->FETCH('syb_more_results');
+	}
 
 	# XXX What to do if $status != 0???
 	
@@ -532,7 +536,7 @@
 
 	$sth->execute || return undef;
 
-	do {
+	{
 	    while(my $d = $sth->fetch) {
 		# The tie() doesn't work here, so call the FETCH method
 		# directly....
@@ -540,13 +544,21 @@
 		   $status = $d->[0];
 	       }
 	    }
-	} while($sth->FETCH('syb_more_results'));
+	    redo if $sth->FETCH('syb_more_results');
+	}
 
 	# XXX What to do if $status != 0???
 	
 	$status;
     }
 	
+#    if($DBI::VERSION >= 1.34) {
+#	*syb_ct_get_data = *ct_get_data;
+#	*syb_ct_data_info = *ct_data_info;
+#	*syb_ct_send_data = *ct_send_data;
+#	*syb_ct_prepare_send = *ct_prepare_send;
+#	*syb_ct_finish_send = *ct_finish_send;
+#    }
 		    
 }
 
@@ -590,7 +602,7 @@ C<BEGIN{}> block:
        $ENV{SYBASE} = '/opt/sybase/11.0.2';
    }
 
-   $dbh = DBI->connect('dbi:Sybase:', $user, $passwd);
+   my $dbh = DBI->connect('dbi:Sybase:', $user, $passwd);
 
 
 =head2 Specifying the server name
@@ -633,7 +645,7 @@ connect to. This will by-pass the server name lookup in the interfaces file.
 This is useful in the case where the server hasn't been entered in the 
 interfaces file.
 
-     $dbh = DBI->connect("dbi:Sybase:host=db1.domain.com;port-4100",
+     $dbh = DBI->connect("dbi:Sybase:host=db1.domain.com;port=4100",
 			 $user, $passwd);
 
 =item maxConnect
@@ -785,6 +797,14 @@ connection negotiation:
 
   $dbh->DBI->connect("dbi:Sybase:sslCAFile=/usr/local/sybase/trusted.txt.ENGINEERING", $user, $password); 
 
+=item bulkLogin
+
+Set this to 1 if the connection is going to be used for a bulk-load
+operation (see I<Experimental Bulk-Load functionality> elsewhere in this
+document.)
+
+  $dbh->DBI->connect("dbi:Sybase:bulkLogin=1", $user, $password); 
+
 =back
 
 These different parameters (as well as the server name) can be strung
@@ -822,11 +842,13 @@ need to re-start the C<fetch()> loop.
 To make sure all results are fetched, the basic C<fetch> loop can be 
 written like this:
 
-     do {
+     {
          while($d = $sth->fetch) {
             ... do something with the data
          }
-     } while($sth->{syb_more_results});
+
+         redo if $sth->{syb_more_results};
+     }
 
 You can get the type of the current result set with 
 $sth->{syb_result_type}. This returns a numerical value, as defined in 
@@ -851,7 +873,7 @@ to your script then you can use the symbolic values (CS_xxx_RESULT)
 instead of the numeric values in your programs, which should make them 
 easier to read.
 
-See also the C<syb_output_param> func() call to handle stored procedures 
+See also the C<$sth->syb_output_params> call to handle stored procedures 
 that B<only> return B<OUTPUT> parameters.
 
 =head1 $sth->execute() failure mode behavior
@@ -1020,7 +1042,8 @@ not supported when using DBD::Sybase to connect to a MS-SQL server.
 =item syb_chained_txn (bool)
 
 If set then we use CHAINED transactions when AutoCommit is off. 
-Otherwise we issue an explicit BEGIN TRAN as needed. The default is off.
+Otherwise we issue an explicit BEGIN TRAN as needed. The default is on
+if it is supported by the server.
 
 This attribute should usually be used only during the connect() call:
 
@@ -1035,6 +1058,9 @@ If set, then identifiers that would normally clash with Sybase reserved
 words can be quoted using C<"identifier">. In this case strings must
 be quoted with the single quote.
 
+This attribute can only be set if the database handle is idle (no
+active statement handle.)
+
 Default is for this attribute to be B<off>.
 
 =item syb_rowcount (int)
@@ -1043,6 +1069,8 @@ Setting this attribute to non-0 will limit the number of rows returned by
 a I<SELECT>, or affected by an I<UPDATE> or I<DELETE> statement to the
 I<rowcount> value. Setting it back to 0 clears the limit.
 
+This attribute can only be set if the database handle is idle.
+
 Default is for this attribute to be B<0>.
 
 =item syb_do_proc_status (bool)
@@ -1050,14 +1078,18 @@ Default is for this attribute to be B<0>.
 Setting this attribute causes $sth->execute() to fetch the return status
 of any executed stored procs in the SQL being executed. If the return
 status is non-0 then $sth->execute() will report that the operation 
-failed (ie it will return C<undef>). This will B<NOT> cause an error
-to be raised if RaiseError is set, however. To get that behaviour
-you need to generate a user error code in the stored proc via
-a 
+failed. 
 
-    raiserror <num> <errmsg> 
+B<NOTE> The result status is NOT the first result set that
+is fetched from a stored proc execution. If the procedure includes
+SELECT statements then these will be fetched first, which means that 
+C<$sth->execute> will NOT return a failure in that case as DBD::Sybase
+won't have seen the result status yet at that point.
 
-statement.
+The RaiseError will NOT be triggered by a non-0 return status if 
+there isn't an associated error message either generated by Sybase
+(duplicate insert error, etc) or generated in the procedure via a T-SQL
+C<raiserror> statement.
 
 Setting this attribute does B<NOT> affect existing $sth handles, only
 those that are created after setting it. To change the behavior of 
@@ -1087,6 +1119,22 @@ For example:
 
 This is very useful information to have when reporting a problem.
 
+=item syb_server_version
+=item syb_server_version_string
+
+These two attributes return the Sybase server version, respectively
+version string, and can be used to turn server-specific functionality
+on or off.
+
+Example:
+
+    print "$dbh->{syb_server_version}\n$dbh->{syb_server_version_string}\n";
+
+prints
+
+    12.5.2
+    Adaptive Server Enterprise/12.5.2/EBF 12061 ESD#2/P/Linux Intel/Enterprise Linux/ase1252/1844/32-bit/OPT/Wed Aug 11 21:36:26 2004
+
 =item syb_failed_db_fatal (bool)
 
 If this is set, then a connect() request where the I<database>
@@ -1110,8 +1158,6 @@ Default: off
 
 =item syb_bind_empty_string_as_null (bool)
 
-B<New in 0.95>
-
 If this attribute is set then an empty string (i.e. "") passed as
 a parameter to an $sth->execute() call will be converted to a NULL
 value. If the attribute is not set then an empty string is converted to
@@ -1121,8 +1167,6 @@ Default: off
 
 =item syb_cancel_request_on_error (bool)
 
-B<New in 0.95>
-
 If this attribute is set then a failure in a multi-statement request
 (for example, a stored procedure execution) will cause $sth->execute()
 to return failure, and will cause any other results from this request to 
@@ -1133,6 +1177,12 @@ up to version 0.94.
 
 Default: on
 
+=item syb_date_fmt (string)
+
+Defines the date/time conversion string when fetching data. See the 
+entry for the C<syb_date_fmt()> method elsewhere in this document for a
+description of the available formats.
+
 =back
 
 =head2 Statement Handle Attributes
@@ -1163,13 +1213,13 @@ when calling $sth->fetch (or any variation).
 
 Instead, you would use
 
-    $sth->func($column, \$data, $size, 'ct_get_data');
+    $sth->syb_ct_get_data($column, \$data, $size);
 
 to retrieve the IMAGE or TEXT data. If $size is 0 then the entire item is
 fetched, otherwis  you can call this in a loop to fetch chunks of data:
 
     while(1) {
-	$sth->func($column, \$data, 1024, 'ct_get_data');
+        $sth->syb_ct_get_data($column, \$data, 1024);
 	last unless $data;
 	print OUT $data;
     }
@@ -1177,8 +1227,10 @@ fetched, otherwis  you can call this in a loop to fetch chunks of data:
 The fetched data is still subject to Sybase's TEXTSIZE option (see the
 SET command in the Sybase reference manual). This can be manipulated with
 DBI's B<LongReadLen> attribute, but C<$dbh->{LongReadLen}> I<must> be 
-set before $sth->execute() is called to take effect (note that LongReadLen
-has no effect  when using DBD::Sybase with an MS-SQL server).
+set before $dbh->prepare() is called to take effect (this is a change
+in 1.05 - previously you could call it after the prepare() but 
+before the execute()). Note that LongReadLen
+has no effect when using DBD::Sybase with an MS-SQL server.
 
 B<Note>: The IMAGE or TEXT column that is to be fetched this way I<must> 
 be I<last> in the select list.
@@ -1193,12 +1245,11 @@ elsewhere in this document.
 
 By default DBD::Sybase will return I<DATETIME> and I<SMALLDATETIME>
 columns in the I<Nov 15 1998 11:13AM> format. This can be changed
-via a special B<_date_fmt()> function that is accessed via the $dbh->func()
-method.
+via a private B<syb_date_fmt()> method.
 
 The syntax is
 
-    $dbh->func($fmt, '_date_fmt');
+    $dbh->syb_date_fmt($fmt);
 
 where $fmt is a string representing the format that you want to apply.
 
@@ -1239,6 +1290,20 @@ Nov 15 1998 11:30AM
 
 11:30:11
 
+=item ISO
+
+2004-08-21 14:36:48.080
+
+=item ISO_strict
+
+2004-08-21T14:36:48.080Z
+
+Note that Sybase has no concept of a timezone, so the trailing "Z" is
+really not correct (assumes that the time is in UTC). However, there
+is no guarantee that the client and the server run in the same timezone,
+so assuming the timezone of the client isn't really a valid option
+either.
+
 =back
 
 =head1 Retrieving OUTPUT parameters from stored procedures
@@ -1260,7 +1325,7 @@ can use this shorthand:
 
     $sth = $dbh->prepare('...');
     $sth->execute;
-    @results = $sth->func('syb_output_params');
+    @results = $sth->syb_output_params();
 
 This will return an array for all the OUTPUT parameters in the proc call,
 and will ignore any other results. The array will be undefined if there are 
@@ -1271,14 +1336,16 @@ The more generic way looks like this:
    $sth = $dbh->prepare("declare \@id_value int, \@id_name
       exec my_proc @name = 'a string', @number = 1234, @id = @id_value OUTPUT, @out_name = @id_name OUTPUT");
    $sth->execute;
-   do {
+   {
       while($d = $sth->fetch) {
          if($sth->{syb_result_type} == 4042) { # it's a PARAM result
             $id_value = $d->[0];
             $id_name  = $d->[1];
          }
       }
-   } while($sth->{syb_more_results});
+
+      redo if $sth->{syb_more_results};
+   }
 
 So the OUTPUT params are returned as one row in a special result set.
 
@@ -1324,7 +1391,16 @@ There is an alternative way to access and update IMAGE/TEXT data
 using the natice OpenClient API. This is done via $h->func() calls,
 and is, unfortunately, a little convoluted.
 
-=head2 Handling IMAGE/TEXT data with ct_get_data()/ct_send_data()
+=head2 Handling IMAGE/TEXT data with syb_ct_get_data()/syb_ct_send_data()
+
+With DBI 1.34 and later you can call all of these ct_xxx() calls directly
+as statement handle methods by prefixing them with syb_, so for example
+
+    $sth->func($col, $dataref, $numbytes, 'ct_fetch_data');
+
+becomes
+
+    $sth->syb_ct_fetch_data($col, $dataref, $numbytes);
 
 =over 4
 
@@ -1348,20 +1424,23 @@ The call sequence is:
     $sth->execute;
     while($d = $sth->fetchrow_arrayref) {
        # The data is in the second column
-       $len = $sth->func(2, \$img, 0, 'ct_get_data');
+       $len = $sth->syb_ct_get_data(2, \$img, 0);
+       # with DBI 1.33 and earlier, this would be
+       # $len = $sth->func(2, \$img, 0, 'ct_get_data');
     }
 
 ct_get_data() returns the number of bytes that were effectively fetched,
 so that when fetching chunks you can do something like this:
 
    while(1) {
-      $len = $sth->func(2, $imgchunk, 1024, 'ct_get_data');
+      $len = $sth->syb_ct_get_data(2, $imgchunk, 1024);
       ... do something with the $imgchunk ...
       last if $len != 1024;
    }
 
 To explain further: Sybase stores IMAGE/TEXT data separately from 
-normal table data, in a chain of 2k blocks. To update an IMAGE/TEXT
+normal table data, in a chain of pagesize blocks (a Sybase database page
+is defined at the server level, and can be 2k, 4k, 8k or 16k in size.) To update an IMAGE/TEXT
 column Sybase needs to find the head of this chain, which is known as
 the "text pointer". As there is no I<where> clause when the ct_send_data()
 API is used we need to retrieve the I<text pointer> for the correct
@@ -1409,24 +1488,25 @@ up via ct_prepare_send() and ct_data_info() for this to work. ct_send_data()
 returns B<TRUE> on success, and B<FALSE> on failure.
 
 In this example, we wish to update the data in the I<img> column
-where the I<id> column is 1:
+where the I<id> column is 1. We assume that DBI is at version 1.34 or
+later and use the direct method calls:
 
   # first we need to find the CS_IODESC data for the data
   $sth = $dbh->prepare("select img from imgtable where id = 1");
   $sth->execute;
   while($sth->fetch) {    # don't care about the data!
-      $sth->func('CS_GET', 1, 'ct_data_info');
+      $sth->syb_ct_data_info('CS_GET', 1);
   }
 
   # OK - we have the CS_IODESC values, so do the update:
-  $sth->func('ct_prepare_send');
+  $sth->syb_ct_prepare_send();
   # Set the size of the new data item (that we are inserting), and make
   # the operation unlogged
-  $sth->func('CS_SET', 1, {total_txtlen => length($image), log_on_update => 0}, 'ct_data_info');
+  $sth->syb_ct_data_info('CS_SET', 1, {total_txtlen => length($image), log_on_update => 0});
   # now transfer the data (in a single chunk, this time)
-  $sth->func($image, length($image), 'ct_send_data');
+  $sth->syb_ct_send_data($image, length($image));
   # commit the operation
-  $sth->func('ct_finish_send');
+  $sth->ct_finish_send();
 
 The ct_send_data() call can also transfer the data in chunks, however you 
 must know the total size of the image before you start the insert. For example:
@@ -1437,14 +1517,14 @@ must know the total size of the image before you start the insert. For example:
   $sth = $dbh->prepare("select img from imgtable where id = 1");
   $sth->execute;
   while($sth->fetch) {    # don't care about the data!
-      $sth->func('CS_GET', 1, 'ct_data_info');
+      $sth->syb_ct_data_info('CS_GET', 1);
   }
 
   # OK - we have the CS_IODESC values, so do the update:
-  $sth->func('ct_prepare_send');
+  $sth->syb_ct_prepare_send();
   # Set the size of the new data item (that we are inserting), and make
   # the operation unlogged
-  $sth->func('CS_SET', 1, {total_txtlen => $size, log_on_update => 0}, 'ct_data_info');
+  $sth->syb_ct_data_info('CS_SET', 1, {total_txtlen => $size, log_on_update => 0});
 
   # open the file, and store it in the db in 1024 byte chunks.
   open(IN, $file) || die "Can't open $file: $!";
@@ -1453,11 +1533,11 @@ must know the total size of the image before you start the insert. For example:
       $bytesread = read(IN, $buff, $to_read);
       $size -= $bytesread;
 
-      $sth->func($buff, $bytesread, 'ct_send_data');
+      $sth->syb_ct_send_data($buff, $bytesread);
   }
   close(IN);
   # commit the operation
-  $sth->func('ct_finish_send');
+  $sth->syb_ct_finish_send();
       
 
 =back
@@ -1488,10 +1568,25 @@ be held on certain system tables for the duration of the transaction.
 If $h->{syb_chained_txn} is I<on>, then DBD::Sybase sets the
 I<CHAINED> option, which tells Sybase not to commit anything automatically.
 Again, you will need to call $dbh->commit() to make any changes to the data
-permanent. In this case Sybase will not let you issue I<BEGIN TRAN> 
-statements in the SQL code that is executed, so if you need to execute
-stored procedures that have I<BEGIN TRAN> statements in them you 
-must use $h->{syb_chained_txn} = 0, or $h->{AutoCommit} = 1.
+permanent. 
+
+=head2 Behavior of $dbh->last_insert_id
+
+This version of DBD::Sybase includes support for the last_insert_id() call,
+with the following caveats:
+
+The last_insert_id() call is simply a wrapper around a "select @@identity"
+query. To be successful (i.e. to return the correct value) this must
+be executed on the same connection as the INSERT that generated the
+new IDENTITY value. Therefore the statement handle that was used to
+perform the insert B<must> have been closed/freed before last_insert_id()
+can be called. Otherwise last_insert_id() will be forced to open a different
+connection to perform the query, and will return an invalid value (usually
+in this case it will return 0).
+
+last_insert_id() ignores any parameters passed to it, and will NOT return
+the last @@identity value generated in the case where placeholders were used,
+or where the insert was encapsulated in a stored procedure.
 
 =head1 Using ? Placeholders & bind parameters to $sth->execute
 
@@ -1635,9 +1730,7 @@ at http://sybooks.sybase.com/
 
 =head1 Stored Procedures and Placeholders
 
-B<NOTE: This feature is experimental>
-
-This version of DBD::Sybase introduces the ability to use ?-style
+DBD::Sybase has the ability to use ?-style
 placeholders as parameters to stored proc calls. The requirements are
 that the stored procedure call be initiated with an "exec" and that it be
 the only statement in the batch that is being prepared():
@@ -1667,7 +1760,7 @@ fetch() and/or $sth->func('syb_output_params'):
 
     my $sth = $dbh->prepare("exec my_proc \@p1 = ?, \@p2 = ?, \@p3 = ? OUTPUT ");
     $sth->execute('one', 'two', 'three');
-    my (@data) = $sth->func('syb_output_params');
+    my (@data) = $sth->syb_output_params();
 
 DBD::Sybase does not attempt to figure out the correct parameter type
 for each parameter (it would be possible to do this for most cases, but
@@ -1685,6 +1778,9 @@ use the explicit call once for each parameter:
     $sth->execute('two', 3.456);
     etc...
 
+Note that once a type has been defined for a parameter you can't change
+it.
+
 When binding SQL_NUMERIC or SQL_DECIMAL data you may get fatal conversion
 errors if the scale or the precision exceeds the size of the target
 parameter definition.
@@ -1710,6 +1806,137 @@ setting the I<arithabort> option:
 See the I<set> command in the Sybase Adaptive Server Enterprise Reference 
 Manual for more information on the set command and on the arithabort option.
 
+=head1 Other Private Methods
+
+=head2 DBD::Sybase private Statement Handle Methods
+
+=over 4
+
+=item @data = $sth->syb_describe([$assoc])
+
+Retrieves the description of each of the output columns of the current 
+result set. Each element of the returned array is a reference
+to a hash that describes the column. The following fields are set:
+NAME, TYPE, SYBTYPE, MAXLENGTH, SCALE, PRECISION, STATUS.
+
+You could use it like this:
+
+   my $sth = $dbh->prepare("select name, uid from sysusers");
+   $sth->execute;
+   my @description = $sth->syb_describe;
+   print "$description[0]->{NAME}\n";         # prints ’name’
+   print "$description[0]->{MAXLENGTH}\n";    # prints 30
+   ....
+
+   while(my $row = $sth->fetch) {
+      ....
+   }
+
+The STATUS field is a string which can be tested for the following
+values: CS_CANBENULL, CS_HIDDEN, CS_IDENTITY, CS_KEY, CS_VERSION_KEY, 
+CS_TIMESTAMP and CS_UPDATABLE. See table 3-46 of the Open Client Client 
+Library Reference Manual for a description of each of these values.
+
+The TYPE field is the data type that Sybase::CTlib converts the
+column to when retrieving the data, so a DATETIME column will be
+returned as a CS_CHAR_TYPE column.
+
+The SYBTYPE field is the real Sybase data type for this column.
+
+I<Note that the symbolic values of the CS_xxx symbols isn't available
+yet in DBD::Sybase.>
+
+
+=back
+
+=head1 Experimental Bulk-Load Functionality
+
+Starting with release 1.04.2 DBD::Sybase has the ability to use Sybase's
+BLK (bulk-loading) API to perform fast data loads. Basic usage is as follows:
+
+  my $dbh = DBI->connect('dbi:Sybase:server=MY_SERVER;bulkLogin=1', $user, $pwd);
+
+  $dbh->begin_work;  # optional.
+  my $sth = $dbh->prepare("insert the_table values(?, ?, ?, ?, ?)",
+                          {syb_bcp_attribs => { identity_flag => 0,
+                                               identity_column => 0 }}});
+  while(<DATA>) {
+    chomp;
+    my @row = split(/\|/, $_);   # assume a pipe-delimited file...
+    $sth->execute(@row);
+  }
+  $dbh->commit;
+  print "Sent ", $sth->rows, " to the server\n";
+  $sth->finish;
+
+First, you need to specify the new I<bulkLogin> attribute in the connection
+string, which turns on the CS_BULK_LOGIN property for the connection. Without
+this property the BLK api will not be functional.
+
+You call $dbh->prepare() with a regular INSERT statement and the 
+special I<syb_bcp_attribs> attribute to turn on BLK handling of the data.
+The I<identity_flag> sub-attribute can be set to 1 if your source data
+includes the values for the target table's IDENTITY column. If the
+target table has an IDENTITY column but you want the insert operation to
+generate a new value for each row then leave I<identity_flag> at 0, but set
+I<identity_col> to the column number of the identity column (it's usually
+the first column in the table, but not always.)
+
+The number of placeholders in the INSERT statement I<must> correspond to
+the number of columns in the table, and the input data I<must> be in the
+same order as the table's physical column order. Any column list in the
+INSERT statement (i.e. I<insert table(a, b, c,...) values(...)> is ignored.
+
+The value of AutoCommit is ignored for BLK operations - rows are only 
+commited when you call $dbh->commit.
+
+You can call $dbh->rollback to cancel any uncommited rows, but this I<also>
+cancels the rest of the BLK operation: any attempt to load rows to the
+server after a call to $dbh->rollback() will fail.
+
+If a row fails to load due to a CLIENT side error (such as a data conversion
+error) then $sth->execute() will return a failure (i.e. false) and
+$sth->errstr will have the reason for the error.
+
+If a row fails on the SERVER side (for example due to a duplicate row
+error) then the entire batch (i.e. between two $dbh->commit() calls) 
+will fail. This is normal behavior for BLK/bcp.
+
+The Bulk-Load API is very sensitive to data conversion issues, as all the
+conversions are handled on the client side, and the row is pre-formatted
+before being sent to the server. By default any conversion that is flagged
+by Sybase's cs_convert() call will result in a failed row. Some of these
+conversion errors are patently fatal (e.g. converting 'Feb 30 2001' to a
+DATETIME value...), while others are debatable (e.g. converting 123.456 to
+a NUMERIC(6,2) which results in a loss of precision). The default behavior
+of failing any row that has a conversion error in it can be modified by 
+using a special error handler. Returning 0 from this handler
+tells DBD::Sybase to fail this row, and returning 1 means that we still
+want to try to send the row to the server (obviously Sybase's internal
+code can still fail the row at that point.)
+You set the handler like this:
+
+    DBD::Sybase::syb_set_cslib_cb(\&handler);
+
+and a sample handler:
+
+   sub cslib_handler {
+     my ($layer, $origin, $severity, $errno, $errmsg, $osmsg, $blkmsg) = @_;
+     
+     print "Layer: $layer, Origin: $origin, Severity: $severity, Error: $errno\n";
+     print $msg;
+     print $osmsg if($osmsg);
+     print $blkmsg if $blkmsg;
+
+     return 1 if($errno == 36)
+
+     return 0;
+   }
+
+Please see the t/xblk.t test script for some examples.
+
+Reminder - this is an I<experimental> implementation. It may change
+in the future, and it could be buggy.
 
 =head1 Using DBD::Sybase with MS-SQL 
 
@@ -1755,7 +1982,7 @@ being returned by the routine (to allow processing of large result sets,
 for example).
 
 C<nsql> also checks three special attributes to enable deadlock retry logic
-\(I<Note> none of these attributes have any effect anywhere else at the moment):
+(I<Note> none of these attributes have any effect anywhere else at the moment):
 
 =over 4
 
